//----------------------------------------------------------------------------------
// File:   MetaBalls.fx
// Author: Yury Uralsky
// Email:  sdkfeedback@nvidia.com
// 
// Copyright (c) 2007 NVIDIA Corporation. All rights reserved.
//
// TO  THE MAXIMUM  EXTENT PERMITTED  BY APPLICABLE  LAW, THIS SOFTWARE  IS PROVIDED
// *AS IS*  AND NVIDIA AND  ITS SUPPLIERS DISCLAIM  ALL WARRANTIES,  EITHER  EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED  TO, IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL  NVIDIA OR ITS SUPPLIERS
// BE  LIABLE  FOR  ANY  SPECIAL,  INCIDENTAL,  INDIRECT,  OR  CONSEQUENTIAL DAMAGES
// WHATSOEVER (INCLUDING, WITHOUT LIMITATION,  DAMAGES FOR LOSS OF BUSINESS PROFITS,
// BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
// ARISING OUT OF THE  USE OF OR INABILITY  TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
// BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
//
//
//----------------------------------------------------------------------------------

Texture3D<float4> volumeDistance;
SamplerState linearSampler : IMMUTABLE
{
    Filter = MIN_MAG_MIP_LINEAR;
    AddressU = Clamp;
    AddressV = Clamp;
	AddressW = Clamp;
};

struct SampleData
{
    float4 Pos : SV_Position;
    float4 Field : TEXCOORD0;   // Gradient in .xyz, value in .w
};

struct SurfaceVertex
{
    float4 Pos : SV_Position;
    float3 N : NORMAL;
};


cbuffer cb0 : register(b0)
{
    row_major float3x3 ViewIT			: VIEW;
    row_major float4x4 	ViewProj			: VIEWPROJECTION;
    row_major float4x4 	World			: WORLD;
	float3 				TetrahedronsFrom;
	float3 				TetrahedronsTo;
};



// Vertex shader calculates field contributions at each grid vertex
SampleData SampleFieldVS(float3 Pos : POSITION)
{
    SampleData o;

    //float3 WorldPos = mul(float4(Pos, 1), World).xyz;

    // Sum up contributions from all metaballs

    o.Field = 0;
	float3 pv = Pos/TetrahedronsTo;
	
	//pv.y = (pv.y-1)*(-1);
	//pv.z = (pv.z-1)*(-1);
	
	o.Field = volumeDistance.SampleLevel(linearSampler,pv,0);
   

    o.Pos = mul(float4((Pos.xyz+TetrahedronsFrom), 1), ViewProj);
    //o.Field.xyz = -normalize(mul(o.Field.xyz, ViewIT));  // we want normals in view space

    // Generate in-out flags

    return o;
}



// Estimate where isosurface intersects grid edge with endpoints v0, v1
SurfaceVertex CalcIntersection(SampleData v0, SampleData v1)
{
    SurfaceVertex o;

	// We're taking special care to generate bit-exact results regardless of traversal (v0,v1) or (v1, v0)

	float t = (2.0 - (v0.Field.w + v1.Field.w)) / (v1.Field.w - v0.Field.w);
    
	o.Pos = 0.5 * (t * (v1.Pos - v0.Pos) + (v1.Pos + v0.Pos));
	o.N = 0.5 * (t * (v1.Field.xyz - v0.Field.xyz) + (v1.Field.xyz + v0.Field.xyz));

    return o;
}

// This struct stores vertex indices of up to 4 edges from the input tetrahedron. The GS code below 
// uses these indices to index into the input vertex set for interpolation along those edges. 
// It basically encodes topology for the output triangle strip (of up to 2 triangles).
struct TetrahedronIndices 
{ 
	uint4 e0; 
	uint4 e1; 
};

// putting edge tables in a tbuffer seems to give better perf
tbuffer EdgeTables
{
    static const TetrahedronIndices EdgeTableGS[16] =
    {
        { 0, 0, 0, 0 }, { 0, 0, 0, 0 },
        { 3, 0, 3, 1 }, { 3, 2, 0, 0 },
        { 2, 1, 2, 0 }, { 2, 3, 0, 0 },
        { 2, 0, 3, 0 }, { 2, 1, 3, 1 },
        { 1, 2, 1, 3 }, { 1, 0, 0, 0 },
        { 1, 0, 1, 2 }, { 3, 0, 3, 2 },
        { 1, 0, 2, 0 }, { 1, 3, 2, 3 },
        { 3, 0, 1, 0 }, { 2, 0, 0, 0 },
        { 0, 2, 0, 1 }, { 0, 3, 0, 0 },
        { 0, 1, 3, 1 }, { 0, 2, 3, 2 },
        { 0, 1, 0, 3 }, { 2, 1, 2, 3 },
        { 3, 1, 2, 1 }, { 0, 1, 0, 0 },
        { 0, 2, 1, 2 }, { 0, 3, 1, 3 },
        { 1, 2, 3, 2 }, { 0, 2, 0, 0 },
        { 0, 3, 2, 3 }, { 1, 3, 0, 0 },
        { 0, 0, 0, 0 }, { 0, 0, 0, 0 },
    };
}

[MaxVertexCount(4)]
void TessellateTetrahedraGS(lineadj SampleData In[4], inout TriangleStream<SurfaceVertex> Stream)
{
    // construct index for this tetrahedron
    uint index = (uint(In[0].Field.w > 1) << 3) |
        (uint(In[1].Field.w > 1) << 2) |
        (uint(In[2].Field.w > 1) << 1) |
         uint(In[3].Field.w > 1);


    // don't bother if all vertices out or all vertices in
    if (index > 0 && index < 15)
    {
        uint4 e0 = EdgeTableGS[index].e0;
        uint4 e1 = EdgeTableGS[index].e1;

        // Emit a triangle
        Stream.Append( CalcIntersection(In[e0.x], In[e0.y]) );
        Stream.Append( CalcIntersection(In[e0.z], In[e0.w]) );
        Stream.Append( CalcIntersection(In[e1.x], In[e1.y]) );

        // Emit additional triangle, if necessary
        if (e1.z != 0) {
            Stream.Append( CalcIntersection(In[e1.z], In[e1.w]) );
        }
    }
}




float4 ShadeSurfacePS( SurfaceVertex IN ) : SV_Target
{


    return float4(1,1,0,.5);
}

DepthStencilState EnableDepthDSS
{
    DepthEnable = true;
    DepthWriteMask = 1;
};

// Tessellate isosurface in a single pass
technique10 MarchingTetrahedronSinglePassGS
{
    pass P0
    {
        //SetRasterizerState( SolidRS );
        SetDepthStencilState( EnableDepthDSS, 0 );

        SetVertexShader( CompileShader( vs_4_0, SampleFieldVS() ) );
        SetGeometryShader( CompileShader( gs_4_0, TessellateTetrahedraGS() ) );
        SetPixelShader( CompileShader( ps_4_0, ShadeSurfacePS() ) );
    }
}